#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <ctype.h>
#include <assert.h>

#include "types.h"

#define true 1
#define false 0

#define EMPTY 0
#define PAWN 1
#define KNIGHT 2
#define BISHOP 3
#define ROOK 4
#define QUEEN 5
#define KING 6

#define BLACK 16

#define row(a) ((a) >> 3)
#define column(a) ((a) & 7)
#define locn(y,x) (((y) << 3) | (x))

/* ------------------------------------------ */

struct MoveNode;

struct PosNode {
    struct MoveNode *parent;
    struct MoveNode *child;
    int hashval;
    int offset;
    int status;
};

#define REMOVED 1
#define VISITED 2
#define INLIST 4

struct MoveNode {
    struct PosNode *parent;
    struct PosNode *child;
    struct MoveNode *next_parent;
    struct MoveNode *next_child;
    int move;
};

struct Position {
    UBYTE square[64];
    int wtm;
};

struct HashTable {
    int size;
    int used;
    struct PosNode **data;
};

/* ------------------------------------------ */

struct PosNode *BuildTree(char *filename);
struct HashTable *MakeHashTable();
void RemoveHashTable(struct HashTable *ht);
int EncodeMove(char *buffer,struct Position *pos);
void HashInsert(struct PosNode *node,struct HashTable *ht);
struct PosNode *FindPosition(struct Position *pos,struct HashTable *ht);
void MakeMove(struct Position *pos,int move);
void InitBoard(struct Position *pos);
int HashBoard(struct Position *pos);
void PrintBoard(struct Position *pos);

struct PosNode **OptimizeTree(struct PosNode *tree,int *nodecnt);
int Equivalent(struct PosNode *a,struct PosNode *b);
int CountNodes(struct PosNode *tree);
void NoneVisited(struct PosNode *tree);
int InsertLeafs(struct PosNode *tree,struct PosNode **list);

void DumpTree(struct PosNode **list,int nodes,char *libfile);

/* ------------------------------------------ */

int main(int argc, char *argv[])
{
    struct PosNode *tree,**list;
    char *infile,*outfile;
    int nodes;

    if (argc != 3) {
	printf("Usage: makebook infile outfile\n");
	exit(1);
    }
    infile = argv[1];
    outfile = argv[2];

    tree = BuildTree(infile);
    list = OptimizeTree(tree,&nodes);

    if (list[nodes-1] != tree) {
	printf("main: Internal error (root not at end of list)\n");
	exit(1);
    }

    DumpTree(list,nodes,outfile);

    return 0;
}

/* --------------------------------------------------------- */

struct PosNode *BuildTree(char *filename)
{
    struct PosNode *root,*parent,*child;
    FILE *infile;
    struct HashTable *ht;
    struct Position *pos;
    char buffer[128];
    int c,move;
    int nummoves = 0;

    if ((infile = fopen(filename,"r")) == NULL) {
	printf("BuildTree: Cannot open file '%s'\n",filename);
	exit(1);
    }

    if ( ((pos = malloc(sizeof(struct Position))) == 0) ||
	 ((ht = MakeHashTable()) == 0) ) {
	printf("BuildTree: Out of memory\n");
	exit(1);
    }

    InitBoard(pos);

    while (fscanf(infile, "%s", &buffer[0]) == 1) {
	if (buffer[0] == '!') {
	    int printit = false;
	    while (((c = fgetc(infile)) != '\n') && (c >= 0)) {
		if (!isspace(c)) {
		    if (!printit)
			printf("  ");
		    printit = true;
		}
		if (printit)
		    putchar(c);
	    }
	    if (printit)
		putchar('\n');
	    InitBoard(pos);
	} else if (buffer[0] == '#') {
	    break;
	} else {
	    struct MoveNode *tmp,*mv;

	    parent = FindPosition(pos,ht);
	    move = EncodeMove(buffer,pos);
	    MakeMove(pos,move);
	    child = FindPosition(pos,ht);
	    for (tmp = parent->child; tmp; tmp = tmp->next_child)
		if (tmp->child == child)
		    goto nextinput;
	    mv = malloc(sizeof(struct MoveNode)); assert(mv);
	    mv->parent = parent;
	    mv->child = child;
	    mv->next_child = parent->child; parent->child = mv;
	    mv->next_parent = child->parent; child->parent = mv;
	    mv->move = move;
	    nummoves++;
	}
    nextinput:;
    }
    printf("\nMoves in tree: %d\n",nummoves);

    fclose(infile);
    InitBoard(pos);
    root = FindPosition(pos,ht);

    RemoveHashTable(ht);
    free(pos);

    return root;
}

struct HashTable *MakeHashTable()
{
    struct HashTable *ht;
    int i;

    if ((ht = malloc(sizeof(struct HashTable))) == 0)
	return 0;
    ht->size = 17;
    ht->used = 0;
    ht->data = malloc(sizeof(void *)*ht->size); assert(ht->data);
    for (i=0; i<ht->size; i++)
	ht->data[i] = 0;

    return ht;
}

void RemoveHashTable(struct HashTable *ht)
{
    free(ht->data);
    free(ht);
}

int EncodeMove(char *buffer,struct Position *pos)
{
    int from,to;
    char c;

    if (strncmp(buffer, "o-o-o", 5) == 0) {
	if (pos->wtm)
	    return (locn(0,4)<<6) + locn(0,2);
	else
	    return (locn(7,4)<<6) + locn(7,2);
    } else if (strncmp(buffer, "o-o", 3) == 0) {
	if (pos->wtm)
	    return (locn(0,4)<<6) + locn(0,6);
	else
	    return (locn(7,4)<<6) + locn(7,6);
    } else {
	c = toupper(buffer[0]);
	if ((c < 'A') || (c > 'H') || (buffer[1] < '1') || (buffer[1] > '8')) {
	    printf("EncodeMove: Illegal move: %s\n",buffer);
	    exit(1);
	}
	from = (c - 'A') + (buffer[1] - '1') * 8;

	c = toupper(buffer[2]);
	if ((c < 'A') || (c > 'H') || (buffer[3] < '1') || (buffer[3] > '8')) {
	    printf("EncodeMove: Illegal move: %s\n",buffer);
	    exit(1);
	}
	to = (c - 'A') + (buffer[3] - '1') * 8;

	return (from << 6) + to;
    }
}

void HashInsert(struct PosNode *node,struct HashTable *ht)
{
    int i;

    for (i = node->hashval % ht->size; ht->data[i]; i = (i+1) % ht->size)
	;
    ht->data[i] = node;
    ht->used++;
}

struct PosNode *FindPosition(struct Position *pos,struct HashTable *ht)
{
    int i,h;
    struct PosNode *node;

    h = HashBoard(pos);

    for (i = h % ht->size; ht->data[i]; i = (i+1) % ht->size)
	if (ht->data[i]->hashval == h)
	    return ht->data[i];

    if (ht->used * 2 >= ht->size) {
	int oldsize;
	struct PosNode **olddata;

	oldsize = ht->size;
	olddata = ht->data;

	ht->size = oldsize * 2 + 1;
	ht->used = 0;
	ht->data = malloc(sizeof(void *)*ht->size); assert(ht->data);
	for (i=0; i<ht->size; i++)
	    ht->data[i] = 0;
	for (i=0; i<oldsize; i++)
	    if (olddata[i])
		HashInsert(olddata[i],ht);

	free(olddata);
    }

    node = malloc(sizeof(struct PosNode)); assert(node);
    node->parent = 0;
    node->child = 0;
    node->hashval = h;
    node->offset = 0;
    node->status = 0;

    HashInsert(node,ht);

    return node;
}

void MakeMove(struct Position *pos,int move)
{
    int from,to;

    from = move >> 6;
    to = move & 0x3f;

    if ( (pos->square[from] == EMPTY) ||
	 ((pos->square[from] ^ ((pos->wtm)<<4)) < BLACK) ||
	 ((pos->square[to] != EMPTY) &&
	  ((pos->square[to] ^ ((pos->wtm)<<4)) >= BLACK)
	  ) ) {
	printf("MakeMove: Illegal move: %03x\n",move);
	exit(1);
    }

    if ((abs(to-from) == 2) && ((pos->square[from] & (BLACK-1)) == KING)) {
	if (pos->wtm) {                                         /* Castling */
	    if (to > from) {
		pos->square[locn(0,7)] = EMPTY;
		pos->square[locn(0,5)] = ROOK;
	    } else {
		pos->square[locn(0,0)] = EMPTY;
		pos->square[locn(0,3)] = ROOK;
	    }
	} else {
	    if (to > from) {
		pos->square[locn(7,7)] = EMPTY;
		pos->square[locn(7,5)] = ROOK | BLACK;
	    } else {
		pos->square[locn(7,0)] = EMPTY;
		pos->square[locn(7,3)] = ROOK | BLACK;
	    }
	}
    } else if ( ((pos->square[from] & (BLACK-1)) == PAWN) &&
		(pos->square[to] == EMPTY) &&
		(column(from) != column(to)) ) {              /* En passant */
	pos->square[locn(row(from),column(to))] = EMPTY;
    }

    pos->square[to] = pos->square[from];
    pos->square[from] = EMPTY;

    if (pos->wtm)
	pos->wtm = false;
    else
	pos->wtm = true;
}

void InitBoard(struct Position *pos)
{
    int i;
    static UBYTE stbrd[64] =
    {
	ROOK,KNIGHT,BISHOP,QUEEN,KING,BISHOP,KNIGHT,ROOK,
	PAWN,PAWN,PAWN,PAWN,PAWN,PAWN,PAWN,PAWN,
	EMPTY,EMPTY,EMPTY,EMPTY,EMPTY,EMPTY,EMPTY,EMPTY,
	EMPTY,EMPTY,EMPTY,EMPTY,EMPTY,EMPTY,EMPTY,EMPTY,
	EMPTY,EMPTY,EMPTY,EMPTY,EMPTY,EMPTY,EMPTY,EMPTY,
	EMPTY,EMPTY,EMPTY,EMPTY,EMPTY,EMPTY,EMPTY,EMPTY,
	PAWN,PAWN,PAWN,PAWN,PAWN,PAWN,PAWN,PAWN,
	ROOK,KNIGHT,BISHOP,QUEEN,KING,BISHOP,KNIGHT,ROOK,
    };

    for (i = 0; i < 64; i++)
	pos->square[i] = ((i >= 48) ? stbrd[i] + BLACK : stbrd[i]);

    pos->wtm = true;
}

int HashBoard(struct Position *pos)
{
    int h = 0;
    int i;

    for (i = 0; i < 64; i++)
	h = h * 37 + pos->square[i];

    if (pos->wtm)
	h += 0x7d843bc9;

    return h & 0x7fffffff;
}

/* --------------------------------------------------------- */

struct PosNode **OptimizeTree(struct PosNode *tree,int *nodecnt)
{
    int i,j,nodes,listsize;
    struct PosNode **list;
    struct MoveNode *tmp,*tmp2;
    int numremoved = 0;

    nodes = CountNodes(tree); NoneVisited(tree);
    list = malloc(sizeof(void *)*nodes); assert(list);
    for (i=0; i<nodes; i++)
	list[i] = 0;

    listsize = InsertLeafs(tree,list); NoneVisited(tree);

    for (i=0; i<nodes; i++) {
	if (i >= listsize) {
	    printf("OptimizeTree: Internal error (cyclic graph)\n");
	    exit(1);
	}
	list[i]->status |= INLIST;
	for (tmp = list[i]->parent; tmp; tmp = tmp->next_parent) {
	    if (tmp->parent->status & INLIST)
		continue;
	    for (tmp2 = tmp->parent->child; tmp2; tmp2 = tmp2->next_child)
		if ((tmp2->child->status & INLIST) == 0)
		    goto nextParent;

	    list[listsize++] = tmp->parent;
	    tmp->parent->status |= INLIST;
	nextParent:;
	}
    }

    for (i=0; i<nodes; i++) {
	if (list[i]->status & REMOVED)
	    continue;
	for (j=i+1; j<nodes; j++) {
	    if (list[j]->status & REMOVED)
		continue;
	    if (Equivalent(list[i],list[j])) {
		list[j]->status |= REMOVED;
		for (tmp = list[j]->parent; tmp; tmp = tmp->next_parent)
		    tmp->child = list[i];

		for (tmp = list[j]->child; tmp; tmp = tmp->next_child)
		    numremoved++;
	    }
	}
    }
    /* From now on, all parent pointers are considered invalid */

    printf("Optimized moves: %d\n",numremoved);
    *nodecnt = nodes;
    return list;
}

int Equivalent(struct PosNode *a,struct PosNode *b)
{
    int a_childs,b_childs;
    struct MoveNode *tmp,*tmp2;

    if (a == b)
	return true;
    else if ((a == 0) || (b == 0))
	return false;

    a_childs = 0;
    for (tmp = a->child; tmp; tmp = tmp->next_child)
	a_childs++;

    b_childs = 0;
    for (tmp = b->child; tmp; tmp = tmp->next_child)
	b_childs++;

    if (a_childs != b_childs)
	return false;

    for (tmp = a->child; tmp; tmp = tmp->next_child) {
	for (tmp2 = b->child; tmp2; tmp2 = tmp2->next_child)
	    if (tmp->move == tmp2->move)
		break;
	if ((tmp2 == NULL) || (tmp->move != tmp2->move))
	    return false;
	if (!Equivalent(tmp->child,tmp2->child))
	    return false;
    }
    return true;
}

int CountNodes(struct PosNode *tree)
{
    int nodes;
    struct MoveNode *tmp;

    if (tree->status & VISITED)
	return 0;
    tree->status |= VISITED;

    if (tree->status & REMOVED)
	nodes = 0;
    else
	nodes = 1;

    for (tmp = tree->child; tmp; tmp = tmp->next_child)
	nodes += CountNodes(tmp->child);

    return nodes;
}

void NoneVisited(struct PosNode *tree)
{
    struct MoveNode *tmp;

    if (tree->status & VISITED) {
	tree->status &= ~VISITED;
	for (tmp = tree->child; tmp; tmp = tmp->next_child)
	    NoneVisited(tmp->child);
    }
}

int InsertLeafs(struct PosNode *tree,struct PosNode **list)
{
    int leafs,l;
    struct MoveNode *tmp;

    if (tree->status & VISITED)
	return 0;
    tree->status |= VISITED;

    if (tree->child == 0) {
	leafs = 1;
	*list++ = tree;
    } else {
	leafs = 0;
    }
    for (tmp = tree->child; tmp; tmp = tmp->next_child) {
	l = InsertLeafs(tmp->child,list);
	leafs += l;
	list += l;
    }
    return leafs;
}

/* --------------------------------------------------------- */

void DumpTree(struct PosNode **list,int nodes,char *outname)
{
    int i,offset,tmpoffs,maxoffs;
    FILE *outfile;
    struct MoveNode *tmp;

    for (i=0; i<nodes-1-i; i++) {
	struct PosNode *tmp = list[i];
	list[i] = list[nodes-1-i];
	list[nodes-1-i] = tmp;
    }

    if ((outfile = fopen(outname,"w")) == NULL) {
	printf("DumpTree: Cannot open file '%s' for writing\n",outname);
	exit(1);
    }

    offset = 0;
    for (i=0; i<nodes; i++) {
	if (list[i]->status & REMOVED)
	    continue;
	list[i]->offset = offset;

	for (tmp = list[i]->child; tmp; tmp = tmp->next_child)
	    offset++;
	offset++;
    }

    fprintf(outfile, "LABEL binbook\n\nASSEMBLE\n");
    fprintf(outfile, "\tCON(5)\t=DOCSTR\n");
    fprintf(outfile, "\tCON(5)\t%d\n", 5+offset*6);

    offset = 0; maxoffs = -1;
    for (i=0; i<nodes; i++) {
	if (list[i]->status & REMOVED)
	    continue;
	for (tmp = list[i]->child; tmp; tmp = tmp->next_child) {
	    fprintf(outfile, "\tCON(3)\t%d\n", tmp->move);
	    tmpoffs = tmp->child->offset - offset;
	    if ((tmpoffs <= 0) || (tmpoffs >= 4095)) {
		printf("DumpTree: Offset too big (%d)\n",tmpoffs);
		exit(1);
	    }
	    if (tmpoffs > maxoffs)
		maxoffs = tmpoffs;
	    fprintf(outfile, "\tCON(3)\t%d\n", tmpoffs);
	    offset++;
	}
	fprintf(outfile, "\tCON(6)\t0\n");
	offset++;
    }
    printf("Max offset: %03x\n",maxoffs);

    fprintf(outfile, "RPL\n");
    fclose(outfile);
}

/* --------------------------------------------------------- */

void PrintBoard(struct Position *pos)
{
    int x, y, p;
    static char outbuf[4096];
    char c;
    UBYTE *board = pos->square;

    outbuf[0] = 0;

    for (y = 7; y >= 0; y--) {
	sprintf(&outbuf[strlen(outbuf)], "\033[31m\033[40m      ");
	for (x = 0; x < 8; x++) {
	    if ((x + y) & 1)
		sprintf(&outbuf[strlen(outbuf)], "\033[40m");
	    else
		sprintf(&outbuf[strlen(outbuf)], "\033[43m");
	    p = board[y * 8 + x];
	    if (p < BLACK) {
		sprintf(&outbuf[strlen(outbuf)], "\033[32m");
	    } else {
		sprintf(&outbuf[strlen(outbuf)], "\033[31m");
		p -= BLACK;
	    }
	    switch (p) {
	    case PAWN   : c = 'b'; break;
	    case KNIGHT : c = 'N'; break;
	    case BISHOP : c = 'B'; break;
	    case ROOK   : c = 'R'; break;
	    case QUEEN  : c = 'Q'; break;
	    case KING   : c = 'K'; break;
	    default     : c = ' '; break;
	    }
	    p = strlen(outbuf);
	    outbuf[p] = ' ';
	    outbuf[p + 1] = c;
	    outbuf[p + 2] = ' ';
	    outbuf[p + 3] = 0;
	}
	sprintf(&outbuf[strlen(outbuf)], "\n");
    }
    sprintf(&outbuf[strlen(outbuf)], "\033[31m\033[40m\n");
    printf("%s", outbuf);
}

